/*
 *  soapUI, copyright (C) 2006 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.mock;

import groovy.lang.Binding;
import groovy.lang.Script;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.activation.DataHandler;
import javax.mail.MessagingException;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.mail.internet.PreencodedMimeBodyPart;
import javax.swing.ImageIcon;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.AttachmentConfig;
import com.eviware.soapui.config.HeaderConfig;
import com.eviware.soapui.config.MockResponseConfig;
import com.eviware.soapui.impl.actions.ShowDesktopPanelAction;
import com.eviware.soapui.impl.wsdl.AbstractWsdlModelItem;
import com.eviware.soapui.impl.wsdl.AttachmentContainer;
import com.eviware.soapui.impl.wsdl.WsdlAttachmentPart;
import com.eviware.soapui.impl.wsdl.WsdlOperation;
import com.eviware.soapui.impl.wsdl.actions.mockresponse.CloneMockResponseAction;
import com.eviware.soapui.impl.wsdl.actions.mockresponse.DeleteMockResponseAction;
import com.eviware.soapui.impl.wsdl.actions.mockresponse.RenameMockResponseAction;
import com.eviware.soapui.impl.wsdl.actions.support.ShowOnlineHelpAction;
import com.eviware.soapui.impl.wsdl.submit.transports.http.AttachmentUtils;
import com.eviware.soapui.impl.wsdl.submit.transports.http.BodyPartAttachment;
import com.eviware.soapui.impl.wsdl.submit.transports.http.MimeMessageMockResponseEntity;
import com.eviware.soapui.impl.wsdl.submit.transports.http.MockResponseDataSource;
import com.eviware.soapui.impl.wsdl.support.FileAttachment;
import com.eviware.soapui.impl.wsdl.support.HelpUrls;
import com.eviware.soapui.impl.wsdl.support.MessageXmlObject;
import com.eviware.soapui.impl.wsdl.support.MessageXmlPart;
import com.eviware.soapui.impl.wsdl.support.MockFileAttachment;
import com.eviware.soapui.impl.wsdl.support.ModelItemIconAnimator;
import com.eviware.soapui.impl.wsdl.support.WsdlAttachment;
import com.eviware.soapui.impl.wsdl.support.soap.SoapVersion;
import com.eviware.soapui.model.iface.Attachment;
import com.eviware.soapui.model.mock.MockResponse;
import com.eviware.soapui.model.mock.MockRunContext;
import com.eviware.soapui.settings.WsdlSettings;
import com.eviware.soapui.support.Tools;
import com.eviware.soapui.support.UISupport;
import com.eviware.soapui.support.action.ActionSupport;
import com.eviware.soapui.support.types.StringToStringMap;

public class WsdlMockResponse extends AbstractWsdlModelItem<MockResponseConfig> implements MockResponse, AttachmentContainer
{
	private final static Log log = LogFactory.getLog( WsdlMockResponse.class );
	
	public final static String MOCKRESULT_PROPERTY = WsdlMockResponse.class.getName() + "@mockresult";
	public final static String SCRIPT_PROPERTY = WsdlMockResponse.class.getName() + "@script";
	public final static String HEADERS_PROPERTY = WsdlMockResponse.class.getName() + "@headers";
	public static final String DISABLE_MULTIPART_ATTACHMENTS = WsdlMockResponse.class.getName() + "@disable-multipart-attachments";
	
	protected List<FileAttachment> attachments = new ArrayList<FileAttachment>();
	private List<WsdlAttachmentPart> definedAttachmentParts;
	private ModelItemIconAnimator iconAnimator;
	private WsdlMockResult mockResult;
	
	public WsdlMockResponse( WsdlMockOperation operation, MockResponseConfig config )
	{
		super( config, operation, "/mockResponse.gif" );
		
		for( AttachmentConfig ac : getConfig().getAttachmentList() )
		{
			attachments.add( new MockFileAttachment( ac, this ));
		}
		
		if( !config.isSetEncoding() )
			config.setEncoding( "UTF-8" );
		
		addAction( new ShowDesktopPanelAction( "Open MockResponse Editor", "Opens the MockResponse Editor for this MockResponse", this ));
      addAction( ActionSupport.SEPARATOR_ACTION );
//      addAction( new AddMockResponseToTestCaseAction( this ));
      addAction( new CloneMockResponseAction( this ));
      addAction( ActionSupport.SEPARATOR_ACTION );
      addAction( new RenameMockResponseAction( this ));
      addAction( new DeleteMockResponseAction( this ));
      addAction( ActionSupport.SEPARATOR_ACTION );
      addAction( new ShowOnlineHelpAction( HelpUrls.MOCKRESPONSE_HELP_URL ));
      
      iconAnimator = new ModelItemIconAnimator( this, "/mockResponse.gif", 
					new String[] {"/exec_request_1.gif", "/exec_request_2.gif",
					"/exec_request_3.gif", "/exec_request_4.gif"} );
	}

	public Attachment[] getAttachments()
	{
		return attachments.toArray( new Attachment[attachments.size()] );
	}
	
	public String getScript()
	{
		return getConfig().isSetScript() ? getConfig().getScript().getStringValue() : null;
	}

	public String getEncoding()
	{
		return getConfig().getEncoding();
	}
	
	public void setEncoding( String encoding )
	{
		String old = getEncoding();
		getConfig().setEncoding( encoding );
		notifyPropertyChanged( ENCODING_PROPERTY, old, encoding );
	}

	public String getResponseContent()
	{
		return getConfig().getResponseContent();
	}

	public void setResponseContent( String responseContent )
	{
		String oldContent = getResponseContent();
		if( !responseContent.equals( oldContent ))
		{
			getConfig().setResponseContent( responseContent );
			notifyPropertyChanged( RESPONSECONTENT_PROPERTY, oldContent, responseContent );
		}
	}
	
	@Override
	public ImageIcon getIcon()
	{
		return iconAnimator.getIcon();
	}

	public WsdlMockOperation getMockOperation()
	{
		return ( WsdlMockOperation ) getParent();
	}
	
	public WsdlMockResult execute( WsdlMockRequest request, WsdlMockResult result ) throws DispatchException
	{
		try
		{
			iconAnimator.start();
			StringToStringMap responseHeaders = getResponseHeaders();
			StringToStringMap contextProperties = request.getContext().toStringToStringMap();
			for( String name : responseHeaders.keySet() )
			{
				result.addHeader( name, Tools.expandProperties( contextProperties, responseHeaders.get( name ), false ));
			}
			
			String responseContent = getResponseContent();
			String script = getScript();
			if( script != null && script.trim().length() > 0 )
			{
				evaluateScript( request );
				contextProperties = request.getContext().toStringToStringMap();
				responseContent = Tools.expandProperties( contextProperties, responseContent, false );
			}
			
			if( !result.isCommitted())
			{
				responseContent = writeResponse( result, responseContent );
			}
			
			result.setReponseContent( responseContent );
			
			setMockResult( result );
			
			return mockResult;
		}
		catch( Throwable e )
		{
			throw new DispatchException( e );
		}
		finally
		{
			iconAnimator.stop();
		}
	}

	public void evaluateScript( WsdlMockRequest request )
	{
		String script = getScript();
		if( script == null || script.trim().length() == 0 )
			return ;
			
		// unsatisfactory way of handling multiple threads -> create pool of scripts instead
		Binding localBinding = new Binding();
		
		WsdlMockService mockService = getMockOperation().getMockService();
		WsdlMockRunner mockRunner = mockService.getMockRunner();
		MockRunContext context = mockRunner == null ? new WsdlMockRunContext( mockService ) : mockRunner.getMockContext(); 
		
		localBinding.setVariable( "context", context );
		localBinding.setVariable( "mockRequest", request);
		localBinding.setVariable( "mockResponse", this );
		localBinding.setVariable( "log", SoapUI.ensureGroovyLog() );
		
		Script localScript = mockService.getScriptShell().parse( script );
		localScript.setBinding( localBinding );
		localScript.run();
	}

	public void setScript( String script )
	{
		String oldScript = getScript();
		if( !script.equals( oldScript ))
		{
			if( !getConfig().isSetScript())
				getConfig().addNewScript();
			getConfig().getScript().setStringValue( script );
			
			notifyPropertyChanged( SCRIPT_PROPERTY, oldScript, script );
		}
	}

	public void setResponseHeaders( StringToStringMap headers )
	{
		StringToStringMap oldHeaders = getResponseHeaders();
		
		HeaderConfig [] headerConfigs = new HeaderConfig[headers.size()];
		int ix = 0;
		for( String header : headers.keySet() )
		{
			headerConfigs[ix] = HeaderConfig.Factory.newInstance();
			headerConfigs[ix].setName( header );
			headerConfigs[ix].setValue( headers.get( header ) );
			ix++;
		}
		
		getConfig().setHeaderArray( headerConfigs );
		
		notifyPropertyChanged( HEADERS_PROPERTY, oldHeaders, headers );
	}

	public StringToStringMap getResponseHeaders()
	{
		StringToStringMap result = new StringToStringMap();
		List<HeaderConfig> headerList = getConfig().getHeaderList();
		for( HeaderConfig header : headerList )
		{
			result.put( header.getName(), header.getValue() );
		}
		
		return result;
	}

	public Attachment attachFile( File file, boolean cache )
	{
		try
		{
			FileAttachment fileAttachment = new MockFileAttachment( file, cache, this );
			attachments.add( fileAttachment );
			notifyPropertyChanged(ATTACHMENTS_PROPERTY, null, fileAttachment );
			return fileAttachment;
		}
		catch (IOException e)
		{
			UISupport.showErrorMessage( e );
			return null;
		}
	}

	public int getAttachmentCount()
	{
		return attachments.size();
	}

	public WsdlAttachment getAttachmentAt( int index )
	{
		return attachments.get( index );
	}

	public void removeAttachment( Attachment attachment )
	{
		int ix = attachments.indexOf( attachment );
		attachments.remove( ix );
		
		try
		{
			notifyPropertyChanged(ATTACHMENTS_PROPERTY, attachment, null );
		}
		finally
		{
			getConfig().removeAttachment( ix);
		}
	}

	public WsdlAttachmentPart[] getDefinedAttachmentParts()
	{
		if( definedAttachmentParts == null )
		{
			try
			{
				WsdlOperation operation = getMockOperation().getOperation();
				if( operation == null )
				{
					definedAttachmentParts = new ArrayList<WsdlAttachmentPart>();
				}
				else
				{
					UISupport.setHourglassCursor();
					definedAttachmentParts = AttachmentUtils.extractAttachmentParts( 
								operation, getResponseContent(), true );
				}
			}
			catch (Exception e)
			{
				log.warn( e.toString() );
			}
			finally 
			{
				UISupport.resetCursor();
			}
		}
		
		return definedAttachmentParts.toArray( new WsdlAttachmentPart[definedAttachmentParts.size()] );
	}

	public WsdlAttachmentPart getAttachmentPart( String partName )
	{
		WsdlAttachmentPart[] parts = getDefinedAttachmentParts();
		for( WsdlAttachmentPart part : parts )
		{
			if( part.getName().equals( partName ))
				return part;
		}
		
		return null;
	}

	public Attachment[] getAttachmentsForPart( String partName )
	{
		List<Attachment> result = new ArrayList<Attachment>();
		
		for( Attachment attachment : attachments )
		{
			if( attachment.getPart().equals( partName ))
				result.add( attachment );
		}
		
		return result.toArray( new Attachment[result.size()]);
	}

	public boolean isMtomEnabled()
	{
		return getSettings().getBoolean( WsdlSettings.ENABLE_MTOM );
	}
	
	public void setMtomEnabled( boolean mtomEnabled )
	{
		boolean old = isMtomEnabled();
		getSettings().setBoolean( WsdlSettings.ENABLE_MTOM, mtomEnabled );
		definedAttachmentParts = null;
		notifyPropertyChanged( MTOM_NABLED_PROPERTY, old, mtomEnabled );
	}

	private String writeResponse(WsdlMockResult response, String responseContent) throws Exception
	{
		MimeMultipart mp = null;
		WsdlOperation operation = getMockOperation().getOperation();
		if( operation == null )
			throw new Exception( "Missing WsdlOperation for mock response");
		
		SoapVersion soapVersion = operation.getInterface().getSoapVersion();
		
		StringToStringMap contentIds = new StringToStringMap();
		boolean isXOP = false;
		
		// preprocess only if neccessary
		if( isMtomEnabled() || getAttachmentCount() > 0 )
		{
			try
			{
				mp = new MimeMultipart();
				
				MessageXmlObject requestXmlObject = new MessageXmlObject(( WsdlOperation ) operation, 
							getResponseContent(), true);
				MessageXmlPart[] requestParts = requestXmlObject.getMessageParts();
				for (MessageXmlPart requestPart : requestParts)
				{
					if (AttachmentUtils.prepareMessagePart(this, mp, requestPart, contentIds))
						isXOP = true;
				}
				responseContent = requestXmlObject.getMessageContent();
			}
			catch (Exception e)
			{
				log.warn( "Failed to process inline/MTOM attachments; " + e );
			}			
		}
		
		response.initResponse();
		
		// non-multipart request?
		if( !isXOP && (mp == null || mp.getCount() == 0 ) && getAttachmentCount() == 0 )
		{
			String encoding = getEncoding();
			byte[] content = encoding == null ? responseContent.getBytes() : responseContent.getBytes(encoding);
			
			response.setContentType( soapVersion.getContentTypeHttpHeader( encoding ));
			response.getOutputStream().write( content );
		}
		else
		{
			// make sure..
			if( mp == null )
				mp = new MimeMultipart();
			
			// init root part
			initRootPart(responseContent, mp, isXOP);
			
			// init mimeparts
			AttachmentUtils.addMimeParts(this, mp, contentIds);
			
			// create request message
			MimeMessage message = new MimeMessage( AttachmentUtils.JAVAMAIL_SESSION );
			message.setContent( mp );
			message.saveChanges();
			MimeMessageMockResponseEntity mimeMessageRequestEntity = new MimeMessageMockResponseEntity( message, isXOP, this );
			
			response.addHeader( "Content-Type", mimeMessageRequestEntity.getContentType() );
			response.addHeader( "MIME-Version", "1.0" );
			mimeMessageRequestEntity.writeRequest( response.getOutputStream() );
		}
		
		return responseContent;
	}
	
	private void initRootPart(String requestContent, MimeMultipart mp, boolean isXOP) throws MessagingException
	{
		MimeBodyPart rootPart = new PreencodedMimeBodyPart( "8bit" );
		rootPart.setContentID( AttachmentUtils.ROOTPART_SOAPUI_ORG  );
		mp.addBodyPart( rootPart, 0 );
		
		DataHandler dataHandler = new DataHandler( new MockResponseDataSource( this, requestContent, isXOP ) );
		rootPart.setDataHandler( dataHandler);
	}

	public Attachment addAttachment( Attachment attachment )
	{
		if( attachment instanceof BodyPartAttachment )
		{
			try
			{
				BodyPartAttachment att = ( BodyPartAttachment ) attachment;
				
				AttachmentConfig newConfig = (AttachmentConfig) getConfig().addNewAttachment();
				newConfig.setData( Tools.readAll( att.getInputStream(), 0 ).toByteArray() );
				newConfig.setContentId( att.getContentID() );
				newConfig.setContentType( att.getContentType() );
				newConfig.setName( att.getName() );
				
				FileAttachment newAttachment = new MockFileAttachment( newConfig, this );
				attachments.add( newAttachment);
				return newAttachment;
			}
			catch( IOException e )
			{
				e.printStackTrace();
			}
		}
		else if( attachment instanceof FileAttachment )
		{
			AttachmentConfig oldConfig = ((FileAttachment)attachment).getConfig();
			AttachmentConfig newConfig = (AttachmentConfig) getConfig().addNewAttachment().set( oldConfig );
			FileAttachment newAttachment = new MockFileAttachment( newConfig, this );
			attachments.add( newAttachment);
			return newAttachment;
		}

		return null;
	}
	
	public void setMockResult( WsdlMockResult mockResult )
	{
		WsdlMockResult oldResult = this.mockResult;
		this.mockResult = mockResult;
		notifyPropertyChanged( MOCKRESULT_PROPERTY, oldResult, mockResult );
	}

	public WsdlMockResult getMockResult()
	{
		return mockResult;
	}
	
	public boolean isMultipartEnabled()
	{
		return !getSettings().getBoolean( DISABLE_MULTIPART_ATTACHMENTS );
	}
	
	public void setMultipartEnabled( boolean multipartEnabled )
	{
		getSettings().setBoolean( DISABLE_MULTIPART_ATTACHMENTS, multipartEnabled );
	}
}
