/*
 *  soapUI, copyright (C) 2006 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.panels.testcase;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.AbstractAction;
import javax.swing.AbstractListModel;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.TestStepConfig;
import com.eviware.soapui.impl.wsdl.testcase.WsdlTestCase;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestStep;
import com.eviware.soapui.impl.wsdl.teststeps.registry.WsdlTestStepFactory;
import com.eviware.soapui.impl.wsdl.teststeps.registry.WsdlTestStepRegistry;
import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.model.support.TestSuiteListenerAdapter;
import com.eviware.soapui.model.testsuite.TestStep;
import com.eviware.soapui.support.UISupport;
import com.eviware.soapui.support.action.ActionList;
import com.eviware.soapui.support.action.ActionSupport;

/**
 * Panel for displaying and editing a list of TestSteps
 * 
 * @author Ole.Matzura
 */

public class TestStepList extends JPanel
{
	private TestStepListModel testStepListModel;
	private JList testStepList;
	private JPopupMenu testListPopup;
	private JMenu appendStepMenu;
	private MoveStepDownAction moveStepDownAction;
	private MoveStepUpAction moveStepUpAction;
	private final WsdlTestCase testCase;

	public TestStepList(WsdlTestCase testCase)
	{
		super(new BorderLayout());
		this.testCase = testCase;

		buildUI();
	}

	private void buildUI()
	{
		testStepListModel = new TestStepListModel();
		testStepList = new JList(testStepListModel);
		testStepList.setCellRenderer(new TestStepCellRenderer());
		testStepList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		testStepList.addKeyListener(new TestStepListKeyHandler());

		testStepList.addMouseListener(new StepListMouseListener());
		testStepList.addMouseListener(new MouseAdapter() 
		{
			private int newIndex;
			private int prevIndex;

			public void mouseClicked(MouseEvent e)
			{
				if( e.getButton() != MouseEvent.BUTTON1 )
					return;
				
				newIndex = testStepList.locationToIndex(e.getPoint());

				if((prevIndex == newIndex) & ( prevIndex != -1))
				{
					testStepList.clearSelection();
					newIndex = -1;
				}

				prevIndex = newIndex;
			}
		} );

		testListPopup = new JPopupMenu();
		testListPopup.addSeparator();

		appendStepMenu = new JMenu("Append Step");

		WsdlTestStepRegistry registry = WsdlTestStepRegistry.getInstance();
		WsdlTestStepFactory[] factories = (WsdlTestStepFactory[]) registry.getFactories();

		for (int c = 0; c < factories.length; c++)
		{
			if (factories[c].canCreate())
				appendStepMenu.add(new InsertTestStepAction(factories[c]));
		}

		testListPopup.add(appendStepMenu);
		
		testListPopup.addSeparator();
		moveStepDownAction = new MoveStepDownAction();
		testListPopup.add(moveStepDownAction);
		moveStepUpAction = new MoveStepUpAction();
		testListPopup.add(moveStepUpAction);

		testListPopup.addPopupMenuListener(new StepListPopupMenuListener(testCase));
		testStepList.setComponentPopupMenu(testListPopup);
		
		add( testStepList, BorderLayout.CENTER );
	}
	
	public void setEnabled(boolean enabled)
	{
		testStepList.setEnabled( enabled );
		
		super.setEnabled(enabled);
	}

	private final class TestStepListKeyHandler extends KeyAdapter
	{
		public void keyPressed(KeyEvent e)
		{
			int ctrlKey = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
			if ((e.getModifiers() & ctrlKey) == ctrlKey && (e.getKeyCode() == KeyEvent.VK_DOWN))
			{
				moveStepDownAction.actionPerformed(new ActionEvent(testStepList, 0, "DOWN"));
				e.consume();
			}
			else if ((e.getModifiers() & ctrlKey) == ctrlKey && (e.getKeyCode() == KeyEvent.VK_UP))
			{
				moveStepUpAction.actionPerformed(new ActionEvent(testStepList, 0, "UP"));
				e.consume();
			}
			else if (e.getKeyChar() == KeyEvent.VK_ENTER)
			{
				int ix = testStepList.getSelectedIndex();
				if (ix != -1)
				{
					TestStep testStep = (TestStep) testCase.getTestStepAt(ix);
					UISupport.selectAndShow(testStep);
					e.consume();
				}
			}
			else
			{
				int ix = testStepList.getSelectedIndex();
				if (ix != -1)
				{
					TestStep testStep = (TestStep) testCase.getTestStepAt(ix);
					ActionList actions = testStep.getActions();
					if( actions != null )
						actions.dispatchKeyEvent( e );
				}
			}
		}
	}

	private final class StepListPopupMenuListener implements PopupMenuListener
	{
		private StepListPopupMenuListener(WsdlTestCase case1)
		{
			super();
		}

		public void popupMenuWillBecomeVisible(PopupMenuEvent e)
		{
			while (testListPopup.getComponentCount() > 4)
				testListPopup.remove(0);

			int ix = testStepList.getSelectedIndex();

			if( SoapUI.getTestMonitor().hasRunningLoadTest( testCase ))
			{
				appendStepMenu.setEnabled( false );
				moveStepDownAction.setEnabled(false);
				moveStepUpAction.setEnabled(false);

				return;
			}

			appendStepMenu.setEnabled(true);
			moveStepUpAction.setEnabled(ix > 0);
			moveStepDownAction.setEnabled(ix > -1 && ix < testCase.getTestStepCount() - 1);

			if( ix >= 0 )
			{
				TestStep testStep = (TestStep) testCase.getTestStepAt(ix);
				testListPopup.add( new JPopupMenu.Separator(), 0  );
				ActionSupport.insertActions(testStep.getActions(), testListPopup, 0);
			}
		}

		public void popupMenuWillBecomeInvisible(PopupMenuEvent e)
		{
		}

		public void popupMenuCanceled(PopupMenuEvent e)
		{
		}
	}

	private final class StepListMouseListener extends MouseAdapter
	{
		public void mouseClicked(MouseEvent e)
		{
			if (e.getClickCount() < 2)
			{
				return;
			}
			
			ModelItem modelItem = (ModelItem) testStepList.getSelectedValue();
			if (modelItem == null)
				return;

			Action defaultAction = modelItem.getActions().getDefaultAction();
			if( defaultAction != null )
				defaultAction.actionPerformed( new ActionEvent( TestStepList.this, 0, null ));
		}
	}

	/**
	 * Renderer which sets icon and wider border for teststeps
	 * 
	 * @author Ole.Matzura
	 */

	private final static class TestStepCellRenderer extends JLabel implements ListCellRenderer
	{
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected,
				boolean cellHasFocus)
		{
			WsdlTestStep testStep = (WsdlTestStep) value;
			setText(testStep.getName());
			setIcon(testStep.getIcon());

			if (isSelected)
			{
				setBackground(list.getSelectionBackground());
				setForeground(list.getSelectionForeground());
			}
			else
			{
				setBackground(list.getBackground());
				setForeground(list.getForeground());
			}

			setEnabled(list.isEnabled());
			setFont(list.getFont());
			setOpaque(true);
			setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));

			return this;
		}
	}

	private class TestStepListModel extends AbstractListModel implements PropertyChangeListener
	{
		private TestStepListTestSuiteListener testStepListTestSuiteListener = new TestStepListTestSuiteListener();

		public TestStepListModel()
		{
			for (int c = 0; c < getSize(); c++)
				testCase.getTestStepAt(c).addPropertyChangeListener(this);

			testCase.getTestSuite().addTestSuiteListener(testStepListTestSuiteListener);
		}

		public int getSize()
		{
			return testCase.getTestStepCount();
		}

		public Object getElementAt(int index)
		{
			return testCase.getTestStepAt(index);
		}

		public void propertyChange(PropertyChangeEvent arg0)
		{
			int ix = testCase.getIndexOfTestStep((TestStep) arg0.getSource());
			if (ix == -1)
				return;

			fireContentsChanged(this, ix, ix);
		}

		public void release()
		{
			testCase.getTestSuite().removeTestSuiteListener(testStepListTestSuiteListener);

			for (int c = 0; c < getSize(); c++)
				testCase.getTestStepAt(c).removePropertyChangeListener(this);
		}

		private class TestStepListTestSuiteListener extends TestSuiteListenerAdapter
		{
			public void testStepAdded(TestStep testStep, int ix)
			{
				if( testStep.getTestCase() == testCase )
				{
					testStep.addPropertyChangeListener(TestStepListModel.this);
					fireIntervalAdded(TestStepListModel.this, ix, ix);
				}
			}

			public void testStepRemoved(TestStep testStep, int ix)
			{
				if( testStep.getTestCase() == testCase )
				{
					testStep.removePropertyChangeListener(TestStepListModel.this);
					fireIntervalRemoved(TestStepListModel.this, ix, ix);
				}
			}
		}
	}

	private class MoveStepDownAction extends AbstractAction
	{
		public MoveStepDownAction()
		{
			super("Move down");
			putValue( Action.ACCELERATOR_KEY, UISupport.getKeyStroke( "menu DOWN" ));
		}

		public void actionPerformed(ActionEvent e)
		{
			int ix = testStepList.getSelectedIndex();
			if (ix == -1 || ix >= testCase.getTestStepCount() - 1)
				return;

			testCase.moveTestStep(ix, 1);
			testStepList.setSelectedIndex(ix+1);
		}
	}

	private class MoveStepUpAction extends AbstractAction
	{
		public MoveStepUpAction()
		{
			super("Move up");
			putValue( Action.ACCELERATOR_KEY, UISupport.getKeyStroke( "menu UP" ));
		}

		public void actionPerformed(ActionEvent e)
		{
			int ix = testStepList.getSelectedIndex();
			if (ix < 1)
				return;

			testCase.moveTestStep(ix, -1);
			testStepList.setSelectedIndex(ix-1);
		}
	}

	public class InsertTestStepAction extends AbstractAction
	{
		private final WsdlTestStepFactory factory;

		public InsertTestStepAction(WsdlTestStepFactory factory)
		{
			super(factory.getTestStepName());
			putValue( Action.SHORT_DESCRIPTION, factory.getTestStepDescription());
			putValue( Action.SMALL_ICON, factory.getTestStepIcon() );
			this.factory = factory;
		}

		public void actionPerformed(ActionEvent e)
		{
			String name = UISupport.prompt( "Specify name for new step", "Insert Step", factory.getTestStepName());
			if( name != null )
			{
				TestStepConfig newTestStepConfig = factory.createNewTestStep(testCase, name);
				if( newTestStepConfig != null )
				{
					WsdlTestStep testStep = testCase.addTestStep(newTestStepConfig);
					UISupport.selectAndShow( testStep );
				}
			}
		}
	}

	public void setSelectedIndex(int i)
	{
		testStepList.setSelectedIndex(i);
	}

	public void setSelectedValue(TestStep testStep, boolean b)
	{
		testStepList.setSelectedValue(testStep, true);
	}

	public void release()
	{
		testStepListModel.release();
	}
}
