/*
 *  soapUI, copyright (C) 2006 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.ui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.ToolTipManager;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.model.project.Project;
import com.eviware.soapui.model.tree.SoapUITreeModel;
import com.eviware.soapui.model.tree.SoapUITreeNode;
import com.eviware.soapui.model.tree.SoapUITreeNodeRenderer;
import com.eviware.soapui.model.tree.nodes.ProjectTreeNode;
import com.eviware.soapui.model.workspace.Workspace;
import com.eviware.soapui.support.action.ActionList;

/**
 * The soapUI navigator tree
 * 
 * @author Ole.Matzura
 */

public class Navigator extends JPanel
{
	private Workspace workspace;
	private JTree mainTree;
	private SoapUITreeModel treeModel;
	private Set<NavigatorListener> listeners = new HashSet<NavigatorListener>();

	public Navigator( Workspace workspace )
	{
		super( new BorderLayout() );
		this.workspace = workspace;
		
		buildUI();
	}
	
	private void buildUI()
	{
		treeModel = new SoapUITreeModel(workspace);
		
      mainTree = new JTree( treeModel );
      mainTree.setRootVisible(true);
      mainTree.setExpandsSelectedPaths(true);
      mainTree.setToggleClickCount( 0 );
      mainTree.addMouseListener(new TreeMouseListener());
      mainTree.addTreeSelectionListener(new InternalTreeSelectionListener());
      mainTree.setCellRenderer( new SoapUITreeNodeRenderer() );
      mainTree.setBorder(null);
      mainTree.getSelectionModel().setSelectionMode( TreeSelectionModel.SINGLE_TREE_SELECTION );
      
      mainTree.addKeyListener( new TreeKeyListener() );
      
      ToolTipManager.sharedInstance().registerComponent(mainTree);
      
      add( new JScrollPane( mainTree ), BorderLayout.CENTER );
	}
	
	public Project getCurrentProject()
   {
      TreePath path = mainTree.getSelectionPath();
      Object node = (Object) path.getLastPathComponent();
      while (node != null && !(node instanceof ProjectTreeNode))
      {
         path = path.getParentPath();
         node = (path == null ? null : path.getLastPathComponent());
      }

      if (node == null)
         return null;

      return ((ProjectTreeNode) node).getProject();
   }
	
	public void addNavigatorListener( NavigatorListener listener )
	{
		listeners.add( listener );
	}
	
	public void removeNavigatorListener( NavigatorListener listener )
	{
		listeners.remove( listener );
	}
	
	public void selectModelItem(ModelItem modelItem)
   {
      TreePath path = treeModel.getPath( modelItem );
      mainTree.setSelectionPath( path );
     	mainTree.expandPath( path );
      mainTree.scrollPathToVisible( path );
   }
	
	private final class TreeKeyListener extends KeyAdapter
	{
		public void keyPressed(KeyEvent e)
		{
			SoapUITreeNode lastPathComponent = (SoapUITreeNode) mainTree.getSelectionPath().getLastPathComponent();
			ActionList actions = lastPathComponent.getModelItem().getActions();
			if( actions != null )
			{
				actions.dispatchKeyEvent( e );
			}
		}
	}
	
	public class InternalTreeSelectionListener implements TreeSelectionListener
   {
      public void valueChanged(TreeSelectionEvent e)
      {
         Object obj = e.getPath().getLastPathComponent();
         if (obj instanceof SoapUITreeNode)
         {
            SoapUITreeNode treeNode = (SoapUITreeNode) obj;

            if( !listeners.isEmpty() )
            {
            	TreePath newPath = e.getNewLeadSelectionPath();
	            NavigatorListener[] array = listeners.toArray( new NavigatorListener[listeners.size()]);
	            for( NavigatorListener listener : array )
	            {
	            	listener.nodeSelected( newPath == null ? null : treeNode );
	            }
            }
         }
      }
   }

	public class TreeMouseListener extends MouseAdapter
   {
      public void mouseClicked(MouseEvent e)
      {
         if (e.isPopupTrigger())
            showPopup(e);
         else if (e.getClickCount() < 2)
            return;
         if (mainTree.getSelectionCount() != 1)
            return;

         TreePath path = mainTree.getSelectionPath();
         SoapUITreeNode node = (SoapUITreeNode) path.getLastPathComponent();
         
         ModelItem modelItem = node.getModelItem();
         ActionList actions = modelItem.getActions();
         if( actions != null )
         	actions.performDefaultAction( new ActionEvent( mainTree, 0, null ));
      }

      public void mousePressed(MouseEvent e)
      {
         if (e.isPopupTrigger())
            showPopup(e);
      }

      public void mouseReleased(MouseEvent e)
      {
         if (e.isPopupTrigger())
            showPopup(e);
      }

      private void showPopup(MouseEvent e)
      {
         TreePath path = mainTree.getPathForLocation((int) e.getPoint().getX(), (int) e.getPoint().getY());
         if (path == null)
            return;
         SoapUITreeNode node = (SoapUITreeNode) path.getLastPathComponent();

         JPopupMenu popupMenu = node.getPopup();
         if (popupMenu == null)
            return;

         mainTree.setSelectionPath(path);

         popupMenu.show(mainTree, e.getX(), e.getY());
      }
   }

	public boolean isVisible( TreePath path )
	{
		return mainTree.isVisible( path );
	}

	public boolean isExpanded( TreePath path )
	{
		return mainTree.isExpanded( path );
	}
}
